#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include <memory.h>
#include <signal.h>
#include <time.h>
#include "PS2Defines.h"
#include "ps2maths.h"
#include "ps2matrix4x4.h"
#include "pad.h"
#include "sps2wrap.h"
#include "dma.h"
#include "texture.h"

// These two external function pointers are exported by the VU code
// and we can now use them as pointers to the compiled VU data
extern "C" void VU_vu1code_start(void);
extern "C" void VU_vu1code_end(void);

bool g_bLoop = true;

void sig_handle(int sig)
{
	g_bLoop = false;
}

void AddCube()
{
	// Front face
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1, 1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 0, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1,-1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 1, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1, 1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 1, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1,-1, 1, 1));	// Vert

	// Right face
	VIFStaticDMA.AddVector(Vector4( 1, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1, 1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 1, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1,-1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 2, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1, 1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 2, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1,-1,-1, 1));	// Vert

	// Back face
	VIFStaticDMA.AddVector(Vector4( 2, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1, 1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 2, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1,-1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 3, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1, 1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 3, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1,-1,-1, 1));	// Vert

	// Left face
	VIFStaticDMA.AddVector(Vector4( 3, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1, 1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 3, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1,-1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 4, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1, 1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 4, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1,-1, 1, 1));	// Vert
}

int main()
{
	// Initialise control pad 0
	if(!pad_init(PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE))
	{
		printf("Failed to initialise control pad\n");
		pad_cleanup(PAD_0);
		exit(0);
	}	

	SPS2Manager.Initialise(4096);	// 4096 * 4K Pages = 16MB Total

	VIFStaticDMA.Initialise(1536);	// 1536 * 4K Pages = 6MB Static DMA
	VIFDynamicDMA.Initialise(256);	// 256 * 4K Pages * 2 Buffers =
									// 2MB Dynamic DMA

	// Load three different textures.
	CTexture Tex1, Tex2, Tex3;
	Tex1.LoadBitmap("test.bmp");
	Tex2.LoadBitmap("test2.bmp");
	Tex3.LoadBitmap("test3.bmp");

	// Two texture buffers, texbuf 0 = 480, texbuf 1 = 496
	int iCurrentTexBuf = 0;

	// Render the box 15 times
	int iNumBoxes = 15;

	// An array of positions of where to draw the box
	Vector4 boxPositions[iNumBoxes];

	//seed the random number generator
  	srand(time(NULL));

	// Generate the random positions for all of the boxes.
	for(int i = 0; i < iNumBoxes; i++)
	{
		boxPositions[i] = Vector4(8 - (rand() % 16), 
								  8 - (rand() % 16),
								  8 - (rand() % 16) -12 , 1);
	}

	// Find out the length of the code in double words
	// We use double words since code is made up of many
	// pairs of two (1 upper, and 1 lower) VU instructions. 
	// Each instruction is 32 bits, so VU code must be a
	// multiple of 64 bits.
	uint32 iCodeLengthInDWORDs = (((uint32)&VU_vu1code_end - (uint32)&VU_vu1code_start) + 7) >> 3;
	// Copy the code to VU1_MICRO_MEM
	memcpy(VU1_MICRO_MEM, &VU_vu1code_start, iCodeLengthInDWORDs * 8);

	// Register our signal function for every possible signal (i.e. ctrl + c)
	for(int sig=0; sig<128; sig++)
		signal(sig, sig_handle);

	// Set up the DMA packet to clear the screen.
	SPS2Manager.InitScreenClear(0x20, 0x05, 0x30);

	// Build a projection matrix
	float n = 2.0f;		// Near plane
	float f = 2000.0f;	// Far plane
	float FovYdiv2 = DegToRad(60.0f) * 0.5f; // 60 degree FOV
	float fAspect = 4.0f / 3.0f;			 // Aspect ratio
	float cotFOV = 1.0f / (Sin(FovYdiv2) / Cos(FovYdiv2));
	// We will be projecting to the 4096 wide drawing area, but we only want
	// the projection matrix to cover the visible area
	float w = cotFOV * (sps2UScreenGetWidth() / 4096.0f) / fAspect;
	float h = cotFOV * (sps2UScreenGetHeight() / 4096.0f);

	Matrix4x4 matProj
		(w, 0, 0, 0,
		 0, -h, 0, 0,
		 0, 0, (f+n) / (f-n), -1,
		 0, 0, (2*f*n) / (f-n), 0);


	Matrix4x4 matView;
	matView.LookAt(Vector4(0, 0, 0, 1), Vector4(0, 0, -10, 1), Vector4(0, 1, 0, 1));

	Matrix4x4 matViewProj = matView * matProj;

	// Now add all of the data that won't change per-frame
	// to the static DMA buffer. The less work the CPU has to do
	// per-frame the better

	// Get the address that we can call to later.
	int iCubeAddr = VIFStaticDMA.GetPointer();

	int iTriangles = 2 + 4 * 3;
	int iVerts = iTriangles + 2;

	// We want to unpack 3 quad words of misc data, and then 3QW per vertex starting at
	// VU mem location 8 (we will upload the matrices to the first 8QW per frame).
	VIFStaticDMA.AddUnpack(V4_32, 8, 3 + iVerts * 3);
	VIFStaticDMA.AddVector(Vector4(2048.0f, 2048.0f, ((float)0xFFFFFF) / 32.0f, *((float *)&iVerts)));

	Vector4 vLight(1.0f, 0.0f, -0.5f, 1.0f);		// A directional light facing into the screen
	vLight.NormaliseSelf();						// Make sure to always use normalised light vectors
	//vLight.DumpVector4("vLight");
	VIFStaticDMA.AddVector(vLight);

	// Add the GIFTag
	VIFStaticDMA.Add128(GS_GIFTAG_BATCH(	iVerts,						// NLOOP
	 										1,							// EOP
	 										1,							// PRE
	 										GS_PRIM(PRIM_TRI_STRIP, 	// PRIM
													PRIM_IIP_GOURAUD, 
													PRIM_TME_ON,
													PRIM_FGE_OFF, 
													PRIM_ABE_ON, 
													PRIM_AA1_OFF, 
													PRIM_FST_STQ, 
													PRIM_CTXT_CONTEXT1, 
													PRIM_FIX_NOFIXDDA),					
	 										GIF_FLG_PACKED,				// FLG
	 										GS_BATCH_3(	GIF_REG_ST,
	 													GIF_REG_RGBAQ, 	// BATCH
	 													GIF_REG_XYZ2)));

	// Add the cube to the static buffer
	AddCube();

	// Flush to make sure VU1 isn't doing anything.
	VIFStaticDMA.Add32(FLUSH);

	// Then run the microcode.
	VIFStaticDMA.Add32(MSCALL(0));

	// Return back to the dynamic chain (note that we aren't aligned on a QWORD boundary, but
	// it doesn't matter because the packet class will sort that out for us).
	VIFStaticDMA.DMARet();

	while(g_bLoop)
	{
		// Fire off the previous frame's buffer (or init data if it's the first time)
		VIFDynamicDMA.Fire();

		pad_update(PAD_0);

		// Set up a world matrix that rotates the cube about it's own axis, and then again
		// around the origin.
		static float sfRot = 0.0f;
		sfRot += 0.03f;

		// Clear the screen
		SPS2Manager.BeginScene();

		Matrix4x4 matWorld, matTrans, matRotY, matWVP;
		matRotY.RotationY(sfRot);

		// Upload the current texture to the free texture buffer
		Tex1.Upload((iCurrentTexBuf == 0) ? 480 :  496);
		iCurrentTexBuf ^= 1;
		Tex1.Select();

		for(int i = 0; i < iNumBoxes; i++)
		{	
			// Change to the second texture
			if(i == 5)
			{
				Tex2.Upload((iCurrentTexBuf == 0) ? 480 :  496);
				iCurrentTexBuf ^= 1;
				Tex2.Select();
			}
			// Then the third
			else if(i == 10)
			{
				Tex3.Upload((iCurrentTexBuf == 0) ? 480 :  496);
				iCurrentTexBuf ^= 1;
				Tex3.Select();
			}

			matTrans.Translation(boxPositions[i].x,
								 boxPositions[i].y,
								 boxPositions[i].z);

			matWorld = matRotY * matTrans;
			matWVP = matWorld * matViewProj;

			// Flush so that we wait for the previous cube to finish rendering
			// before we upload new data into that area of memory.
			VIFDynamicDMA.Add32(FLUSH);

			VIFDynamicDMA.AddUnpack(V4_32, 0, 8);
			VIFDynamicDMA.AddMatrix(matWVP);
			VIFDynamicDMA.AddMatrix(matWorld);
			VIFDynamicDMA.DMACall(iCubeAddr);
		}

		SPS2Manager.EndScene();
		
		// Check for exit condition
		if((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)) g_bLoop = false;
	}
	
	pad_cleanup(PAD_0);

	return 0;
}

