; These are equates, they work exactly like #define in C
; We use them here for the locations of the various data in VU memory
ProjMat			.equ  	0
WorldMat		.equ    4
Scale			.equ	8
NumVertsMem		.equ	8
LightMem		.equ    9
GIFTag			.equ    10
StartSTQ		.equ    11
StartNorm		.equ    12
StartVert		.equ	13

; This is a file full of macros that comes with VCL
.include "vcl_sml.i"

; Tell VCL that we want to be able to use all vf, and vi registers
.init_vf_all
.init_vi_all

.syntax new
.vu

; This is the entry point, and is where execution will start on VU1
--enter
--endenter

	fcset		0x000000					; VCL won't let us use CLIP without first zeroing
											; the clip flags

	iaddiu		iVert, vi00, 0				; Start vertex counter
	iaddiu		iVertPtr, vi00, 0			; Point to the first vert

	ilw.w		iNumVerts, NumVertsMem(vi00); We now load the number of verts from memory
											; instead of using a .equ
	
	lq			fScales, Scale(vi00)		; A scale vector that we will add
											; to and scale the vertices by after
											; projecting them
											
	lq			fLight, LightMem(vi00)		; Load the light direction
	sub.xyz		fLight, vf00, fLight		; Reverse the light vector so that we can
											; dot product it with the normals to give
											; the light's intensity
	
	; Load the transformation matrix
	MatrixLoad  fTransform, ProjMat, vi00
	
	; Load the world matrix
	MatrixLoad  fWorldMat, WorldMat, vi00
	
loop:										; For each vertex
	lq			Vert, StartVert(iVertPtr)	; Load the vector (currently in object space
											; and floating point format)
	
	; Transform the vertex by the projection matrix we passed in (this is a macro from
	; vcl_sml.i)
	MatrixMultiplyVertex Vert, fTransform, Vert
	
	; *** NEW ***
	; CLIPPING
	clipw.xyz	Vert, Vert					; This instruction checks if the vertex is outside
											; the viewing frustum. If it is, then the appropriate
											; clipping flags are set
	fcand		vi01, 0x3FFFF				; Bitwise AND the clipping flags with 0x3FFFF, this makes
											; sure that we get the clipping judgement for the last three
											; verts (i.e. that make up the triangle we are about to draw)
	iaddiu		iADC, vi01, 0x7FFF			; Add 0x7FFF. If any of the clipping flags were set this will
											; cause the triangle not to be drawn (any values above 0x8000
											; that are stored in the w component of XYZ2 will set the ADC
											; bit, which tells the GS not to perform a drawing kick on this
											; triangle. 
	isw.w		iADC, StartVert(iVertPtr)	; Finally store the ADC bit back to memory
	; END CLIPPING
    
    div         q,    vf00[w], Vert[w]		; Work out 1.0f / w
    mul.xyz     Vert, Vert, q				; Project the vertex by dividing by W
    
    lq			STQ,  StartSTQ(iVertPtr)	; Load the texture coordinates
    mul			STQ,  STQ, q				; Divide by W
    sq			STQ,  StartSTQ(iVertPtr)	; Store the texture coordinates back
    
    lq			Normal, StartNorm(iVertPtr) ; Load the vertex normal
    
	; Transform the normal into world space
    MatrixMultiplyVertex Normal, fWorldMat, Normal
    
    VectorDotProduct Normal, fLight, Normal ; Puts the light intensity into Normal.x
    
    loi			0.1							; Ambient
	maxi.x		Normal, Normal, i			; if( Normal.x < Ambient ) Normal.x = Ambient;
	
	loi			128							
	addi.xyz	FinalCol, vf00, i				; Light colour (0x80, 0x80, 0x80)
	mul.xyz		FinalCol, FinalCol,	Normal[x]	; Scale by the intensity
	addi.w		FinalCol, vf00, i
	
	ftoi0		FinalCol, FinalCol				; Convert to integer
	sq			FinalCol, StartNorm(iVertPtr)	; Then write back to memory	
    
    mula.xyz	acc, fScales, vf00[w]		; Move fScales into the accumulator (acc = fScales * 1.0)
 	madd.xyz	Vert, Vert, fScales			; Multiply and add the scales (Vert = Vert * fScales + fScales)
	ftoi4.xyz	Vert, Vert					; Convert the vertex to 12:4 fixed point for the GS
	
	sq.xyz		Vert, StartVert(iVertPtr)	; Write the vertex back to VU memory

	iaddiu		iVert, iVert, 1				; Increment the vertex counter
	iaddiu		iVertPtr, iVertPtr, 3		; Increment the vertex pointer
	ibne		iVert, iNumVerts, loop		; Branch back to "loop" label if
											; iVert and iNumVerts are not equal
											
	iaddiu		iGIFTag, vi00, GIFTag		; Load the position of the giftag
	xgkick		iGIFTag						; and tell the PS2 to send that to the GS
--exit
--endexit